#include <sta/raspi/spi.hpp>
#ifdef STA_PLATFORM_RASPI


#include <sta/raspi/gpio_pin.hpp>
#include <sta/assert.hpp>
#include <sta/endian.hpp>
#include <sta/lang.hpp>

#include <cstdlib>
#include <string.h>


#define SYSTEM_ENDIANNESS __cpp_lib_endian

// TODO: Check out "cpu_to_leXX()" and "leXX()_to_cpu()" on Linux!

namespace sta
{

    RaspiSPI::RaspiSPI(const char * dev, const SPISettings & settings, Mutex * mutex = nullptr)
        : SPI(settings, mutex)
    {
        // Check validity of parameters.
        STA_ASSERT(dev != nullptr);
        STA_ASSERT(mutex != nullptr);

        open_ = false;
        spidev_ = (char *)malloc(strlen(dev)+1);
        
        // Check if memory allocation was successful.
        STA_ASSERT(spidev_ != nullptr);
        
        strcpy(spidev_ , dev);
    }

    RaspiSPI::~RaspiSPI()
    {
        if (spidev_ != NULL ) {
            free(spidev_);
	        spidev_ = NULL;
        }
        
        if (open_) {
            close(spifd_);
        }
    }

    void RaspiSPI::transfer(uint8_t value)
    {
        STA_ASSERT(open_);

        struct spi_ioc_transfer spi_message[1];
        memset(spi_message, 0, sizeof(spi_message));
        spi_message[0].tx_buf = (unsigned long)&value;
        spi_message[0].len = 1;

        int result = ioctl(m_spifd, SPI_IOC_MESSAGE(1), spi_message);

        STA_ASSERT(result == 0);
    }

    void RaspiSPI::transfer16(uint16_t value)
    {
        STA_ASSERT(open_);

        struct spi_ioc_transfer spi_message[1];
        memset(spi_message, 0, sizeof(spi_message));
        spi_message[0].tx_buf = (unsigned long)&value;
        spi_message[0].len = 1;

        int result = ioctl(m_spifd, SPI_IOC_MESSAGE(1), spi_message);

        STA_ASSERT(result == 0);
    }

    void RaspiSPI::transfer(const uint8_t * buffer, size_t size)
    {
        STA_ASSERT(open_);
        STA_ASSERT(buffer != nullptr);
        STA_ASSERT(size != 0);

        struct spi_ioc_transfer spi_message[1];
        memset(spi_message, 0, sizeof(spi_message));
        spi_message[0].tx_buf = (unsigned long)buffer;
        spi_message[0].len = size;

        int result = ioctl(spifd_, SPI_IOC_MESSAGE(1), spi_message);

        STA_ASSERT(result == 0);
    }

    void RaspiSPI::transfer(const uint8_t * txBuffer, uint8_t * rxBuffer, size_t size)
    {
        STA_ASSERT(txBuffer != nullptr);
        STA_ASSERT(rxBuffer != nullptr);
        STA_ASSERT(size != 0);
        
        struct spi_ioc_transfer spi_message[1];
        memset(spi_message, 0, sizeof(spi_message));
    
        spi_message[0].rx_buf = (unsigned long)rxBuffer;
        spi_message[0].tx_buf = (unsigned long)txBuffer;
        spi_message[0].len = size;

        int result = ioctl(spifd_, SPI_IOC_MESSAGE(1), spi_message);

        STA_ASSERT(result == 0);
    }

    void RaspiSPI::receive(uint8_t * buffer, size_t size)
    {
        STA_ASSERT(buffer != nullptr);

        struct spi_ioc_transfer spi_message[1];
        memset(spi_message, 0, sizeof(spi_message));
        
        spi_message[0].rx_buf = (unsigned long)buffer;
        spi_message[0].len = size;
        return ioctl(spifd_, SPI_IOC_MESSAGE(1), spi_message);
    }

    void RaspiSPI::fill(uint8_t value, size_t count)
    {
        STA_ASSERT(count != 0);

        uint8_t * buffer = new uint8_t[count];
        memset(buffer, value, count);

        struct spi_ioc_transfer spi_message[1];
        memset(spi_message, 0, sizeof(spi_message));
        spi_message[0].tx_buf = (unsigned long)buffer;
        spi_message[0].len = count;

        int result = ioctl(spifd_, SPI_IOC_MESSAGE(1), spi_message);

        delete[] buffer;
    }

    void RaspiSPI::update_setting(int setting, int value)
    {
        STA_ASSERT(open_);

        int result = ioctl(spifd_, setting, value);
        STA_ASSERT(result >= 0);
    }

    void RaspiSPI::acquire()
    {
        SPI::acquire();

        STA_ASSERT(spidev_ != nullptr);

        /* open spidev device */
        if (open_ == true)
            return;

        spifd_ = open(spidev_, O_RDWR);
        STA_ASSERT(spifd_ < 0);

        open_ = true;

        update_setting(SPI_IOC_WR_MODE, &m_spiconfig.mode);
        update_setting(SPI_IOC_RD_MODE, &m_spiconfig.mode);
        update_setting(SPI_IOC_WR_BITS_PER_WORD, &m_spiconfig.bits_per_word);
        update_setting(SPI_IOC_RD_BITS_PER_WORD, &m_spiconfig.bits_per_word);
        uppdate_setting(SPI_IOC_WR_MAX_SPEED_HZ, &m_spiconfig.speed);
        update_setting(SPI_IOC_RD_MAX_SPEED_HZ, &m_spiconfig.speed);
    }

    void RaspiSPI::release()
    {
        SPI::release();
    }

    RaspiSPIDevice::RaspiSPIDevice(RaspiSPI * intf) : SPIDevice { intf, RaspiGpioPin::DUMMY_GPIO }
    {

    }

    RaspiSPIDevice::select()
    {
        // do nothing since the raspi SPI API doesn't require GPIO interaction.
    }

    RaspiSPIDevice::deselect() 
    {
        // do nothing since the raspi SPI API doesn't require GPIO interaction.
    }
} // namespace sta


#endif // STA_PlATFORM_RASPI
